"""ASAI can find the available actions, conditions, and other useful IAM information for AWS services.

This was written to make building detailed IAM policies with regional permissions boundaries less tedious.

## Features

* Search or list all AWS services
* List all regional or global AWS services
* Create a list of available IAM actions, wildcarding common prefixes
* Display all available IAM actions for a service
* Print all IAM information for a service.
"""
__author__ = """Shaun Martin"""
__email__ = "inhumantsar@protonmail.com"
__version__ = "0.3.0"

from . import models
from ._backend import (
    get_actions,
    get_global_services,
    get_policies,
    get_regional_services,
    get_service_by_prefix,
    get_services,
    search_services,
)

__all__ = [
    "models",
    "get_actions",
    "get_policies",
    "get_service_by_prefix",
    "get_services",
    "get_global_services",
    "get_regional_services",
    "search_services",
]



"""Main module."""

import json
import logging
import re
from typing import Dict, Generator, List, Tuple, Union

import requests
from fuzzywuzzy import fuzz

from asai.models import AWSService

log = logging.getLogger()

_POLICIES_URL = "https://awspolicygen.s3.amazonaws.com/js/policies.js"
# policies.js sets one big var, trim that var def and we have valid JSON
_POLICIES_PREFIX = "app.PolicyEditorConfig="


def get_policies() -> Dict:
    """Download all possible IAM actions, conditions, etc."""
    resp = requests.get(_POLICIES_URL)
    if resp.status_code != 200:
        log.error(f"Bad response from {_POLICIES_URL}")
        return

    resp_json = resp.content[len(_POLICIES_PREFIX) :]
    return json.loads(resp_json)


def get_service_by_prefix(
    prefix: str, service_list: List[AWSService] = None
) -> Union[AWSService, None]:
    """Find a single service by its StringPrefix. Returns None if no matches are found."""
    services = service_list or get_services()
    try:
        return next(s for s in services if s.StringPrefix == prefix)
    except StopIteration as e:
        return


def get_global_services(service_list: List[AWSService] = None) -> List[AWSService]:
    """List all services which aren't tied to specific regions."""
    services = service_list or get_services()
    return [s for s in services if not s.is_regional]


def get_regional_services(service_list: List[AWSService] = None) -> List[AWSService]:
    """List all services which are tied to specific regions."""
    services = service_list or get_services()
    return [s for s in services if s.is_regional]


def get_services() -> List[AWSService]:
    """Parse out all services from policies."""
    service_map = get_policies()["serviceMap"]
    return [AWSService(name=k, **v) for k, v in service_map.items()]


def get_actions() -> Dict[str, List[str]]:
    """Parse out all service prefixes and actions from policies."""
    services = get_services()
    return {v.StringPrefix: v.Actions for _, v in services.items()}


def search_services(
    search_term: str, min_confidence=70, service_list: List[AWSService] = None
) -> Generator[Tuple[int, AWSService], None, None]:
    """Use fuzzy matching to search the list of service names and prefixes."""
    services = service_list or get_services()
    search_term = search_term.lower()
    for service in get_services():
        name_words = service.name.lower().split()
        name_score = max([fuzz.ratio(i, search_term) for i in name_words])
        prefix_score = fuzz.ratio(service.StringPrefix, search_term)
        score = max(name_score, prefix_score)
        if score > min_confidence:
            yield (score, service)



"""JSON-friendly models representing AWS resources."""

import json
import random
import re
import string
from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Union


def _randomness():
    return "".join(random.sample(string.ascii_letters, 5))


@dataclass(order=True)
class AWSService:
    """Service information as received from policies.js."""

    name: str
    # matching JSON keys for the rest because i'm lazy
    StringPrefix: str
    Actions: List[str]
    HasResource: bool
    ARNFormat: str = field(default="")
    ARNRegex: str = field(default="")
    conditionKeys: List[str] = field(default_factory=lambda: [])

    _CAMEL_REGEX = re.compile(r".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)")

    @property
    def is_regional(self) -> bool:
        """Return true if a region field is present in the service's ARN format."""
        # region field can appear as ${region}, <region>, or region, in either title or lower case
        return "region" in self.ARNFormat.lower()

    @property
    def action_wildcards(self) -> List[str]:
        """Return service actions, wildcarding common prefixes."""
        actions = []
        for action in self.Actions:
            prefix = AWSService._get_action_prefix(action)
            wildcard = f"{prefix}*"
            if wildcard in actions:
                continue

            if AWSService._count_prefix_occurances(prefix, self.Actions) > 1:
                actions.append(wildcard)
            else:
                actions.append(action)
        return sorted(actions)

    def _get_action_prefix(action: str) -> str:
        """Split action by CamelCase, return prefix."""
        matches = re.findall(AWSService._CAMEL_REGEX, action)
        return matches[0]  # .group(0)

    def _count_prefix_occurances(prefix: str, actions: List[str]) -> int:
        """Get the number of times a prefix occurs within a list of strings."""
        return len([i for i, x in enumerate(actions) if x.startswith(prefix)])


@dataclass()
class AWSPolicyStatement:
    """Representation of an IAM policy statement."""

    Sid: str
    Action: List[str]
    Resource: Union[List[str], str]
    Condition: Dict[str, Dict[str, str]]

    def from_services(services: List[AWSService], wildcard=False):
        """Create object using service information, optionally wildcarding actions."""
        sid = f"Sid-{_randomness()}" if len(services) > 1 else services[0].name
        actions, resources = [], []
        conditions_list = []
        for s in services:
            for a in s.action_wildcards if wildcard else s.Actions:
                actions.append(f"{s.StringPrefix}:{a}")
            if s.HasResource:
                resources.append(s.ARNFormat)
            for i in s.conditionKeys:
                if i not in conditions_list:
                    conditions_list.append(i)

        return AWSPolicyStatement(
            Sid=sid,
            Action=sorted(set(actions)),
            Resource=sorted(set(resources)),
            Condition={
                f"SomeCondition-{_randomness()}": {i: "...some value..."}
                for i in sorted(set(conditions_list))
            },
        )

    @property
    def json(self) -> str:
        """Prettified JSON representation of the policy statement."""
        return json.dumps(self.dict, indent=4, sort_keys=False)

    @property
    def dict(self) -> Dict[str, Dict[str, Any]]:
        """Return dictionary with falsey values excluded."""
        statement_dict = {"Sid": self.Sid, "Effect": "Allow", "Action": self.Action}
        if self.Resource:
            statement_dict["Resource"] = self.Resource
        if self.Condition:
            statement_dict["Condition"] = self.Condition
        return statement_dict


@dataclass()
class AWSPolicy:
    """IAM policy document scaffold."""

    Statement: List[AWSPolicyStatement]
    Version: str = "2012-10-17"

    def from_services(services: List[AWSService], wildcard=False):
        """Create object using service information, optionally wildcarding actions."""
        statements = [AWSPolicyStatement.from_services(services, wildcard)]
        return AWSPolicy(Statement=statements)

    @property
    def json(self) -> str:
        """Prettified JSON representation of the policy."""
        d = {
            "Version": self.Version,
            "Statement": [s.dict for s in self.Statement],
        }
        return json.dumps(d, indent=4, sort_keys=False)

    @property
    def dict(self) -> Dict[str, Dict[str, Any]]:
        """Create IAM policy statement scaffold."""
        return asdict(self)



#!/usr/bin/env python

"""Tests for `asai` package."""
import json
from os.path import dirname, join
from pathlib import Path
from typing import List, Tuple

from click.testing import CliRunner

from asai import _cli as cli


def _read(fname):
    """Read content of a file and return as a string."""
    return Path(join(dirname(__file__), "expected", fname)).read_text()


def test_cli_root():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.cli)
    assert result.exit_code == 0
    help_result = runner.invoke(cli.cli, ["--help"])
    assert help_result.exit_code == 0
    assert "--help  Show this message and exit." in help_result.output


def test_cli_services_list():
    types = ["all", "global", "no-arn", "regional"]
    expected = {i: _read(f"services-{i}") for i in types}
    runner = CliRunner()
    result = runner.invoke(cli.cli, args=["services"])
    assert result.exit_code == 0
    assert all([i in result.output for i in list(expected.keys())])
    for cmd, services in expected.items():
        result = runner.invoke(cli.cli, args=["services", cmd])
        assert result.exit_code == 0
        assert all([i in result.output for i in services])


def test_cli_services_search():
    runner = CliRunner()
    searches = {
        "auto": [
            "AWS Auto Scaling (autoscaling-plans)",
            "Amazon EC2 Auto Scaling (autoscaling)",
            "Application Auto Scaling (application-autoscaling)",
        ],
        "ec": [
            "Amazon EC2 Auto Scaling (autoscaling)",
            "Amazon EC2 (ec2)",
            "Amazon EC2 Instance Connect (ec2-instance-connect)",
            "Amazon Elastic Container Service (ecs)",
            "Amazon Elastic Container Registry (ecr)",
            "Amazon EC2 Image Builder (imagebuilder)",
        ],
        "container": [
            "Amazon Elastic Container Registry Public (ecr-public)",
            "Amazon Elastic Container Service (ecs)",
            "Amazon Elastic Container Registry (ecr)",
            "Amazon EMR on EKS (EMR Containers) (emr-containers)",
        ],
    }
    for term, results in searches.items():
        result = runner.invoke(cli.cli, args=["services", "search", term])
        assert result.exit_code == 0
        assert all([i in result.output for i in results])


def test_cli_single_service():
    types = ["actions", "actions-wildcard", "arn-format", "arn-regex", "info"]
    services = ["ec2", "s3", "dynamodb"]
    expected = {f"{i}_{j}": _read(f"service-{i}-{j}") for i in types for j in services}
    runner = CliRunner()
    result = runner.invoke(cli.cli, args=["service"])
    assert result.exit_code == 0
    assert all([i in result.output for i in types])
    for cmd, expected_results in expected.items():
        print(["service", *cmd.split("-")])
        result = runner.invoke(cli.cli, args=["service", *cmd.split("_")])
        assert result.exit_code == 0
        assert all([i in result.output for i in expected_results])


def test_cli_policy():
    runner = CliRunner()
    params = [
        "-p",
        "--prefix",
        "--all-global",
        "--all-regional",
        "-s",
        "--search",
        "--wildcard",
        "--group",
    ]
    result = runner.invoke(cli.cli, args=["policy", "--help"])
    assert result.exit_code == 0
    assert all([i in result.output for i in params])
    commands: List[Tuple[str, str]] = [
        ("all", ""),
        ("all-group", "--group"),
        ("prefixes", "-p ec2 -p ecs"),
        ("prefixes-dedupe", "-p sqs -p sqs"),
        ("prefixes-group", "-p ec2 -p ecs --group"),
        ("searches", "-s container -s access"),
        ("searches-group", "-s container -s access --group"),
        ("searches-prefixes-group", "-s container -s access -p ecs -p sqs --group"),
    ]
    expected = {i[0]: json.loads(_read(f"policy-{i[0]}")) for i in commands}
    for (k, cmd) in commands:
        args = ["policy", *cmd.split()]
        print(args)
        result = runner.invoke(cli.cli, args=args)
        output = json.loads(result.output)
        assert expected[k]["Version"] == output["Version"]
        estatements = expected[k]["Statement"]
        for statement in output["Statement"]:
            assert any([statement["Action"] == e["Action"] for e in estatements])
            if "Resource" in statement.keys():
                assert any(
                    [
                        statement["Resource"] == e["Resource"]
                        for e in estatements
                        if "Resource" in e.keys()
                    ]
                )
            if "Condition" in statement.keys():
                econs = [e["Condition"] for e in estatements if "Condition" in e.keys()]
                out_keys = [
                    list(v.keys())[0] for _, v in statement["Condition"].items()
                ]
                assert any(
                    [
                        sorted(out_keys)
                        == sorted([list(v.keys())[0] for _, v in e.items()])
                        for e in econs
                    ]
                )



"""Console script for asai."""
import sys
from pprint import pprint
from types import FunctionType
from typing import List

import click

import asai
from asai.models import AWSPolicy, AWSPolicyStatement, AWSService


@click.group()
def cli():
    """ASAI can find the available actions, conditions, and other useful IAM information for AWS services."""
    pass


@cli.command("policy")
@click.option(
    "-p",
    "--prefix",
    multiple=True,
    default=[],
    help="Specific service prefix. Can be used multiple times.",
)
@click.option(
    "--all-global",
    is_flag=True,
    help="Include all global services in policy scaffold.",
)
@click.option(
    "--all-regional",
    is_flag=True,
    help="Include all global services in policy scaffold.",
)
@click.option(
    "-s",
    "--search",
    multiple=True,
    default=[],
    help="Include all search results. Can be used multiple times.",
)
@click.option(
    "--wildcard",
    is_flag=True,
    help="Wildcard the actions list of all specified services.",
)
@click.option(
    "--group",
    is_flag=True,
    help="Group services by parameter.",
)
def policy(
    prefix: List[str],
    all_global: bool,
    all_regional: bool,
    search: List[str],
    wildcard: bool,
    group: bool,
):
    """
    Generate an IAM policy scaffold using specified services.

    When multiple service options are specified, they will be combined into a
    single policy.

    If `--group` is set, then the policy will contain multiple statements for each
    parameter. eg: Setting `--prefix ssm --prefix ec2 --all-global --search
    container` will result in four statements, one for SSM, one for EC2, one for
    all global services, and one for "container" search results. These will
    *not* be de-duped.

    If no service options are specified, a policy scaffold with all services
    will be created. Using `--group` here will result in each service getting its own
    policy statement.
    """
    all_services = asai.get_services()
    services = []
    add_to_services: FunctionType = services.append if group else services.extend

    for p in prefix:
        add_to_services([asai.get_service_by_prefix(p, service_list=all_services)])
    if all_global:
        add_to_services(asai.get_global_services(service_list=all_services))
    if all_regional:
        add_to_services(asai.get_regional_services(service_list=all_services))
    for s in search:
        add_to_services(
            [i[1] for i in asai.search_services(s, service_list=all_services)]
        )

    if not services:
        services = [[s] for s in all_services] if group else all_services

    if group:
        statements = [AWSPolicyStatement.from_services(s, wildcard) for s in services]
        click.echo(AWSPolicy(statements).json)
    else:
        click.echo(AWSPolicy.from_services(services, wildcard).json)


@cli.group("services")
def aws_services():
    """List or search for services by attribute."""
    pass


@aws_services.command("all")
def all_services():
    """List all service prefixes."""
    _echo_service_prefixes(asai.get_services())


@aws_services.command()
def no_arn():
    """List all services which don't have an ARN format."""
    services = asai.get_services()
    _echo_service_prefixes([s for s in services if not s.ARNFormat])


@aws_services.command("global")
def global_services():
    """List all services which aren't tied to specific regions."""
    _echo_service_prefixes(asai.get_global_services())


@aws_services.command()
def regional():
    """List all services which are tied to specific regions."""
    _echo_service_prefixes(asai.get_regional_services())


@aws_services.command()
@click.argument("search_term")
def search(search_term: str):
    """Search for a service by name or prefix."""
    results = asai.search_services(search_term)
    for _, svc in sorted(results, reverse=True, key=lambda x: x[0]):
        click.echo(f"{svc.name} ({svc.StringPrefix})")


@cli.group()
def service():
    """Get information on a single service."""
    pass


@service.command()
@click.argument("prefix")
def arn_format(prefix: str):
    """Display a service's ARN format description.

    PREFIX is a service's short name. eg: ec2, ssm, or s3.
    """
    service = _get_service_by_prefix(prefix)
    click.echo(service.ARNFormat)


@service.command()
@click.argument("prefix")
def arn_regex(prefix: str):
    """Display a service's ARN Regex string.

    PREFIX is a service's short name. eg: ec2, ssm, or s3.
    """
    service = _get_service_by_prefix(prefix)
    click.echo(service.ARNRegex)


@service.command()
@click.argument("prefix")
def actions(prefix: str):
    """List a service's actions.

    PREFIX is a service's short name. eg: ec2, ssm, or s3.
    """
    service = _get_service_by_prefix(prefix)
    for action in sorted(service.Actions):
        click.echo(f"{prefix}:{action}")


@service.command()
@click.argument("prefix")
def actions_wildcard(prefix: str):  # noqa: D301
    """List a service's actions, wildcarding common prefixes.

    \b
        $ asai service actions-wildcard s3
        s3:AbortMultipartUpload
        s3:BypassGovernanceRetention
        s3:Create*
        s3:Delete*
        s3:DescribeJob
        s3:Get*
        s3:List*
        s3:ObjectOwnerOverrideToBucketOwner
        s3:Put*
        s3:Replicate*
        s3:RestoreObject
        s3:Update*

    PREFIX is a service's short name. eg: ec2, ssm, or s3.
    """
    service = _get_service_by_prefix(prefix)
    for action in service.action_wildcards:
        click.echo(f"{prefix}:{action}")


@service.command()
@click.argument("prefix")
def info(prefix: str):
    """Print all available service information.

    PREFIX is a service's short name. eg: ec2, ssm, or s3.
    """
    service = _get_service_by_prefix(prefix)

    click.echo(f"{service.name} ({service.StringPrefix})\n")
    click.echo(f"Has Resource: {service.HasResource}")
    click.echo("\nActions:")
    for action in sorted(service.Actions):
        click.echo(f"  - {action}")
    click.echo(f"\nARN Format: {service.ARNFormat}")
    click.echo(f"ARN Regex: {service.ARNRegex}")
    click.echo("\nCondition Keys:")
    for condition_key in service.conditionKeys:
        click.echo(f"  - {condition_key}")


def _echo_service_prefixes(services: List[AWSService]):
    svc_ids = sorted(set([s.StringPrefix for s in services]))
    for svc in svc_ids:
        click.echo(svc)


def _get_service_by_prefix(prefix: str) -> AWSService:
    service = asai.get_service_by_prefix(prefix)
    if not service:
        click.echo(f"{prefix} was not found in the service list.", err=True)
        sys.exit(1)
    return service


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



"""ASAI can find the available actions, conditions, and other useful IAM information for AWS services."""

from os.path import dirname, join
from pathlib import Path

from setuptools import find_packages, setup

GITHUB_URL = "https://github.com/inhumantsar/asai"
VERSION = "0.3.0"


def read(fname):
    """Read content of a file and return as a string."""
    return Path(join(dirname(__file__), fname)).read_text()


def get_requirements():
    """Return requirements with loose version restrictions."""
    return read("requirements.txt").replace("==", ">=").split("\n")


setup(
    name="asai",
    version=VERSION,
    license="BSD License",
    url=GITHUB_URL,
    download_url=f"{GITHUB_URL}/archive/refs/tags/v{VERSION}.zip",
    project_urls={"Repository": GITHUB_URL, "Bug Reports": f"{GITHUB_URL}/issues"},
    author="Shaun Martin",
    author_email="inhumantsar@protonmail.com",
    description="ASAI can find the available actions, conditions, and other useful IAM information for AWS services.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["tests", "tests.*"]),
    include_package_data=True,
    zip_safe=False,
    platforms="any",
    install_requires=get_requirements(),
    test_suite="tests",
    keywords=["aws", "iam", "amazon web services"],
    setup_requires=["pytest-runner"],
    tests_require=["tox==3.24.0", "pytest==6.2.4", "pytest-xdist==2.3.0"],
    python_requires=">=3.6.0",
    entry_points={
        "console_scripts": [
            "asai=asai._cli:cli",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Utilities",
    ],
)
